﻿using BuzzGUI.Common;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Interop;
using System.Windows.Media;
using WDE.ModernSequenceEditor;
using Application = System.Windows.Application;

namespace WDE.ModernSequenceEditor
{   
    public class CustomSequencerWindow : NativeWindow, IDisposable
    {
        public SequenceEditor SequenceEditor { get; internal set; }
        public Window SequenceEditorWindow { get; }
        MainWindowObserver mainWindowObserver;
        ClassicSeqenceEditorObserver classicSeqenceEditorObserver;

        public static bool OneInstanceCreated = false;
        public CustomSequencerWindow(SequenceEditor seq)
        {

            SequenceEditor = seq;

            // This observes main Buzz window move, resize...
            mainWindowObserver = new MainWindowObserver();
            mainWindowObserver.PropertyChanged += MainWindowObserver_PropertyChanged;

            SequenceEditorWindow = new Window();
            SequenceEditorWindow.WindowStyle = WindowStyle.None;
            SequenceEditorWindow.ShowInTaskbar = false;
            SequenceEditorWindow.BorderThickness = new Thickness(0);
            SequenceEditorWindow.AllowsTransparency = false;
            SequenceEditorWindow.ResizeMode = ResizeMode.NoResize;
            SequenceEditorWindow.Effect = null;
            //SequenceEditorWindow.Topmost = true;
            SequenceEditorWindow.MinHeight = 40;
            SequenceEditorWindow.MinWidth = 40;
            

            // This listens to Classic Sequencer Editor hwnd events
            IntPtr buzzClassicSequenceEditorHwnd = FindClassicSequenceEditor(); //FindBuzzMDIClient();
            if (buzzClassicSequenceEditorHwnd != IntPtr.Zero)
            {   
                this.AssignHandle(buzzClassicSequenceEditorHwnd);
                //SendMessage(buzzClassicSequenceEditorHwnd, WM_PAINT, true, 0); // Force paint.
                classicSeqenceEditorObserver = new ClassicSeqenceEditorObserver();
                classicSeqenceEditorObserver.PropertyChanged += MainWindowObserver_PropertyChanged;
            }

            Global.Buzz.PropertyChanged += Buzz_PropertyChanged;

            //if (Global.Buzz.ActiveView == BuzzGUI.Interfaces.BuzzView.SequenceView)
            {
                //UpdateView();
                //SequenceEditorWindow.InvalidateVisual();
            }

            SequenceEditorWindow.Content = seq;
            seq.SetVisibility(true);

            OneInstanceCreated = true;
        }

        private void SetWindowZOrder()
        {
            WindowInteropHelper wif = new WindowInteropHelper(SequenceEditorWindow);

            IntPtr buzzOldSeq = FindClassicSequenceEditor();
            wif.Owner = buzzOldSeq;
            SetParent(wif.Handle, buzzOldSeq);
            //uint flags = (uint)(SetWindowPosFlags.NOMOVE | SetWindowPosFlags.NOSIZE);
            //SetWindowPos(wif.Handle, buzzOldSeq, 0, 0, 0, 0, flags);
        }

        private void MainWindowObserver_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            RECT wRect;
            GetWindowRect(Handle, out wRect);
            UpdateWindow(wRect);
        }

        private void Buzz_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "ActiveView")
            {
                UpdateView();
            }
        }

        public void UpdateView()
        {
            if (Global.Buzz.ActiveView != BuzzGUI.Interfaces.BuzzView.SequenceView)
            {
                SequenceEditorWindow.Visibility = Visibility.Hidden;
            }
            else if (Global.Buzz.ActiveView == BuzzGUI.Interfaces.BuzzView.SequenceView)
            {
                RECT wRect;
                GetWindowRect(Handle, out wRect);
                UpdateWindow(wRect);
                if (!SequenceEditorWindow.IsActive)
                {
                    SetWindowZOrder();
                    SequenceEditorWindow.Show();
                    //InvalidateRect(FindClassicSequenceEditor(), IntPtr.Zero, false);
                }

                SequenceEditorWindow.Visibility = Visibility.Visible;
                SequenceEditorWindow.Focus();
                SequenceEditor.SetVisibility(true);
                SequenceEditor.Focus();
            }
        }

        [System.Security.Permissions.PermissionSet(System.Security.Permissions.SecurityAction.Demand, Name = "FullTrust")]
        protected override void WndProc(ref Message m)
        {
            // Listen for operating system messages
            if (Global.Buzz.ActiveView == BuzzGUI.Interfaces.BuzzView.SequenceView)
            {
                if (m.Msg == WM_PAINT)
                {
                    base.WndProc(ref m);
                }
                else if (m.Msg == WM_PRINTCLIENT)
                {
                    base.WndProc(ref m);
                }
                else if (m.Msg == WS_EX_COMPOSITED)
                {
                    base.WndProc(ref m);
                }
                else if (m.Msg == WM_SETFOCUS)
                {
                    base.WndProc(ref m);
                }
                else if (m.Msg == WM_NCACTIVATE)
                {
                    base.WndProc(ref m);
                }
                else
                {
                    //Global.Buzz.DCWriteLine("WM: " + m.Msg.ToString("X"));
                    base.WndProc(ref m);
                }
            }
            else
            {
                base.WndProc(ref m);
            }
            
        }

        private void UpdateWindow(RECT wRect)
        {   
            UpdateWindow(0, 0, Math.Max(0, wRect.Right - wRect.Left - SystemParameters.VerticalScrollBarWidth), Math.Max(0, wRect.Bottom - wRect.Top - 2 * SystemParameters.HorizontalScrollBarHeight));
        }

        private void UpdateWindow(double top, double left, double width, double height)
        {
            if (Global.Buzz.ActiveView != BuzzGUI.Interfaces.BuzzView.SequenceView)
            {
                SequenceEditorWindow.Visibility = Visibility.Collapsed;
                return;
            }
            
            if (!SequenceEditorWindow.IsActive)
            {
                SequenceEditorWindow.Show();
                SetWindowZOrder();
            }

            SequenceEditorWindow.Visibility = Visibility.Visible;

            WindowInteropHelper wif = new WindowInteropHelper(SequenceEditorWindow);

            SetWindowPos(wif.Handle, IntPtr.Zero, (int)top, (int)left, (int)width, (int)height, 0);
        }

        public IntPtr FindBuzzMDIClient()
        {
            IntPtr result;
            IntPtr hWndParent = GetBuzzWindow();
            result = EnumAllWindows(hWndParent, "MDIClient").FirstOrDefault();


            return result;
        }

        public static IntPtr FindClassicSequenceEditor()
        {
            IntPtr result;
            IntPtr hWndParent = GetBuzzWindow();
            result = EnumAllWindowCaptions(hWndParent, "Sequence Editor").FirstOrDefault();

            return result;
        }

        public void Dispose()
        {
            SequenceEditorWindow.Close();
            SequenceEditor = null;
            this.ReleaseHandle();
            Global.Buzz.PropertyChanged -= Buzz_PropertyChanged;
            mainWindowObserver.PropertyChanged -= MainWindowObserver_PropertyChanged;
            mainWindowObserver.ReleaseHandle();
            if (classicSeqenceEditorObserver != null)
            {
                classicSeqenceEditorObserver.PropertyChanged -= MainWindowObserver_PropertyChanged;
                classicSeqenceEditorObserver.ReleaseHandle();
                classicSeqenceEditorObserver.Dispose();
            }
            OneInstanceCreated = false;
        }


        public static IntPtr GetBuzzWindow()
        {
            // Beautiful!
            IntPtr buzzWnd = Global.Buzz.MachineViewHWND;
            //Global.Buzz.DCWriteLine("HWND Caption: " + GetWinCaption(buzzWnd));
            buzzWnd = GetParent(buzzWnd);
            //Global.Buzz.DCWriteLine("HWND Caption: " + GetWinCaption(buzzWnd));
            buzzWnd = GetParent(buzzWnd);
            //Global.Buzz.DCWriteLine("HWND Caption: " + GetWinCaption(buzzWnd));
            buzzWnd = GetParent(buzzWnd);
            //Global.Buzz.DCWriteLine("HWND Caption: " + GetWinCaption(buzzWnd));

            return buzzWnd;
        }

        class MainWindowObserver : NativeWindow, INotifyPropertyChanged
        {
            public MainWindowObserver()
            {
                IntPtr buzzWnd = FindClassicSequenceEditor();//GetBuzzWindow();

                if (buzzWnd != IntPtr.Zero)
                {
                    this.AssignHandle(buzzWnd);
                }
            }

            [System.Security.Permissions.PermissionSet(System.Security.Permissions.SecurityAction.Demand, Name = "FullTrust")]
            protected override void WndProc(ref Message m)
            {
                // Listen for operating system messages

                if (m.Msg == WM_MOVE ||
                    m.Msg == WM_SIZE)
                {
                    //base.WndProc(ref m);
                    PropertyChanged?.Raise(this, "Location");
                }
                else
                {
                    //Global.Buzz.DCWriteLine("WM Window: " + m.Msg);
                }
                base.WndProc(ref m);
            }

            public event PropertyChangedEventHandler PropertyChanged;
        }

        class ClassicSeqenceEditorObserver : NativeWindow, INotifyPropertyChanged, IDisposable
        {
            public ClassicSeqenceEditorObserver()
            {
                IntPtr seqWnd = FindClassicSequenceEditor();

                if (seqWnd != IntPtr.Zero)
                {
                    this.AssignHandle(seqWnd);
                    SendMessage(GetChildWindows(seqWnd)[0], WM_SETREDRAW, false, 0); // Disble window updates
                }
            }

            [System.Security.Permissions.PermissionSet(System.Security.Permissions.SecurityAction.Demand, Name = "FullTrust")]
            protected override void WndProc(ref Message m)
            {
                if (m.Msg == WM_PAINT || m.Msg == WM_PRINTCLIENT )
                {
                    PropertyChanged?.Raise(this, "Seq Paint");
                }
                if (m.Msg == WM_SETREDRAW) // Testtest
                {
                    base.WndProc(ref m);
                }
                else
                {
                    //Global.Buzz.DCWriteLine("WM Window: " + m.Msg);
                    base.WndProc(ref m);
                }
            }

            public void Dispose()
            {
                IntPtr seqWnd = FindClassicSequenceEditor();
                SendMessage(GetChildWindows(seqWnd)[0], WM_SETREDRAW, true, 0); // Enalbe window updates
            }

            public event PropertyChangedEventHandler PropertyChanged;
        }

        #region Parameters

        public const int WM_MOVE = 0x3;
        public const int WM_SIZE = 0x5;
        public const int WM_SETFOCUS = 0x7;
        public const int WM_NCACTIVATE = 0x86;
        public const int WM_MDIACTIVATE = 0x222;
        public const int WM_MDIDEACTIVATE = 0x229;
        public const int WM_PAINT = 0xf;
        public const int WM_PRINTCLIENT = 0x318;
        public const int WS_EX_COMPOSITED = 0x2000000;

        [System.Runtime.InteropServices.StructLayout(System.Runtime.InteropServices.LayoutKind.Sequential, CharSet = System.Runtime.InteropServices.CharSet.Auto)]
        public struct RECT
        {
            public int Left;
            public int Top;
            public int Right;
            public int Bottom;
        }

        [System.Runtime.InteropServices.StructLayout(System.Runtime.InteropServices.LayoutKind.Sequential, CharSet = System.Runtime.InteropServices.CharSet.Auto)]
        public struct PAINTSTRUCT
        {   
            public IntPtr hdc;
            public bool fErase;
            public RECT rcPaint;
            public bool fRestore;
            public bool fIncUpdate;
        }

        #endregion

        #region user32_calls_and_structs
        // Stuff

        [DllImport("user32.dll", CharSet = CharSet.Auto, ExactSpelling = false)]
        internal static extern IntPtr BeginPaint(HandleRef hWnd, [In][Out] ref PAINTSTRUCT lpPaint);

        [DllImport("user32.dll", CharSet = CharSet.Auto, ExactSpelling = false)]
        internal static extern bool EndPaint(HandleRef hWnd, ref PAINTSTRUCT lpPaint);

        [DllImport("user32.dll", ExactSpelling = true, CharSet = CharSet.Auto)]
        public static extern IntPtr GetParent(IntPtr hWnd);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        public static extern IntPtr SetParent(IntPtr hWndChild, IntPtr hWndNewParent);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern bool SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter, int X, int Y, int cx, int cy, uint uFlags);

        static readonly IntPtr HWND_TOPMOST = new IntPtr(-1);
        static readonly IntPtr HWND_NOTOPMOST = new IntPtr(-2);
        static readonly IntPtr HWND_TOP = new IntPtr(0);
        static readonly IntPtr HWND_BOTTOM = new IntPtr(1);

        /// <summary>
        /// Window handles (HWND) used for hWndInsertAfter
        /// </summary>
        public static class HWND
        {
            public static IntPtr
            NoTopMost = new IntPtr(-2),
            TopMost = new IntPtr(-1),
            Top = new IntPtr(0),
            Bottom = new IntPtr(1);
        }

        /// <summary>
        /// SetWindowPos Flags
        /// </summary>
        public struct SetWindowPosFlags
        {
            public static readonly int
            NOSIZE = 0x0001,
            NOMOVE = 0x0002,
            NOZORDER = 0x0004,
            NOREDRAW = 0x0008,
            NOACTIVATE = 0x0010,
            DRAWFRAME = 0x0020,
            FRAMECHANGED = 0x0020,
            SHOWWINDOW = 0x0040,
            HIDEWINDOW = 0x0080,
            NOCOPYBITS = 0x0100,
            NOOWNERZORDER = 0x0200,
            NOREPOSITION = 0x0200,
            NOSENDCHANGING = 0x0400,
            DEFERERASE = 0x2000,
            ASYNCWINDOWPOS = 0x4000;
        }

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        public static extern int SendMessage(IntPtr hWnd, Int32 wMsg, bool wParam, Int32 lParam);
        private const int WM_SETREDRAW = 11;

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        static extern bool ShowWindow(IntPtr hWnd, int nCmdShow);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        static extern bool InvalidateRect(IntPtr hWnd, IntPtr lpRect, bool bErase);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern bool GetWindowRect(IntPtr hwnd, out RECT lpRect);

        public delegate bool Win32Callback(IntPtr hwnd, IntPtr lParam);
        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        private static extern int GetWindowText(IntPtr hWnd, StringBuilder strText, int maxCount);


        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool EnumChildWindows(IntPtr parentHandle, Win32Callback callback, IntPtr lParam);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        static public extern IntPtr GetClassName(IntPtr hWnd, System.Text.StringBuilder lpClassName, int nMaxCount);

        private static bool EnumWindow(IntPtr handle, IntPtr pointer)
        {
            GCHandle gch = GCHandle.FromIntPtr(pointer);
            List<IntPtr> list = gch.Target as List<IntPtr>;
            if (list == null)
                throw new InvalidCastException("GCHandle Target could not be cast as List<IntPtr>");
            list.Add(handle);
            return true;
        }

        public static List<IntPtr> GetChildWindows(IntPtr parent)
        {
            List<IntPtr> result = new List<IntPtr>();
            GCHandle listHandle = GCHandle.Alloc(result);
            try
            {
                Win32Callback childProc = new Win32Callback(EnumWindow);
                EnumChildWindows(parent, childProc, GCHandle.ToIntPtr(listHandle));
            }
            finally
            {
                if (listHandle.IsAllocated)
                    listHandle.Free();
            }
            return result;
        }

        public static string GetWinClass(IntPtr hwnd)
        {
            if (hwnd == IntPtr.Zero)
                return null;
            StringBuilder classname = new StringBuilder(100);
            IntPtr result = GetClassName(hwnd, classname, classname.Capacity);
            if (result != IntPtr.Zero)
                return classname.ToString();
            return null;
        }

        public static string GetWinCaption(IntPtr hwnd)
        {
            if (hwnd == IntPtr.Zero)
                return null;
            StringBuilder classname = new StringBuilder(100);
            int result = GetWindowText(hwnd, classname, classname.Capacity);
            if (result > 0)
                return classname.ToString();
   
            return null;
        }

        public static IEnumerable<IntPtr> EnumAllWindows(IntPtr hwnd, string childClassName)
        {
            List<IntPtr> children = GetChildWindows(hwnd);
            if (children == null)
                yield break;
            foreach (IntPtr child in children)
            {
                if (GetWinClass(child) == childClassName)
                    yield return child;
                foreach (var childchild in EnumAllWindows(child, childClassName))
                    yield return childchild;
            }
        }

        public static IEnumerable<IntPtr> EnumAllWindowCaptions(IntPtr hwnd, string childClassName)
        {
            List<IntPtr> children = GetChildWindows(hwnd);
            if (children == null)
                yield break;
            foreach (IntPtr child in children)
            {
                if (GetWinCaption(child) == childClassName)
                    yield return child;
                foreach (var childchild in EnumAllWindowCaptions(child, childClassName))
                    yield return childchild;
            }
        }

        #endregion
    }
}
